from antelope import CONTEXT_STATUS_, EntityNotFound, comp_dir  # , BackgroundRequired
from ..interfaces.iforeground import AntelopeForegroundInterface
from antelope_core.implementations import BasicImplementation

from antelope_core.entities.quantities import new_quantity
from antelope_core.entities.flows import new_flow
from ..entities.fragments import InvalidParentChild
from ..entities.fragment_editor import create_fragment, clone_fragment, _fork_fragment  # interpose,


class NotForeground(Exception):
    pass


class UnknownRefQuantity(Exception):
    pass


class AntelopeForegroundImplementation(BasicImplementation, AntelopeForegroundInterface):
    """
    A foreground manager allows a user to build foregrounds.  This should work with catalog references, rather
    than actual entities.

    This interface should cover all operations for web tool

    To build a fragment, we need:
     * child flows lambda
     * uuid

     * parent or None

     * Flow and Direction (relative to parent)
     * cached exchange value (or 1.0)

     * external_ref is uuid always? unless set-- ForegroundArchive gets a mapping from external_ref to uuid

     * other properties: Name, StageName

    To terminate a fragment we need a catalog ref: origin + ref for:
     * to foreground -> terminate to self
     * to subfragment -> just give uuid
     * background terminations are terminations to background fragments, as at present
     * process from archive, to become sub-fragment
     * flow, to become fg emission

    Fragment characteristics:
     * background flag
     * balance flag

    Scenario variants:
     - exchange value
     - termination
    """
    #_count = 0
    #_frags_with_flow = defaultdict(set)  # we actually want this to be shared among
    #_recursion_check = None

    def __init__(self, *args, **kwargs):
        super(AntelopeForegroundImplementation, self).__init__(*args, **kwargs)
        self._observations = []

    '''
    Add some useful functions from other interfaces to the foreground
    '''
    def get_local(self, external_ref, **kwargs):
        """
        The special characteristic of a foreground is its access to the catalog-- so-- use it
        lookup locally; fallback to catalog query- should make origin a kwarg
        :param external_ref:
        :param kwargs:
        :return:
        """
        e = self._fetch(external_ref, **kwargs)
        if e is not None:
            return e
        try:
            origin, external_ref = external_ref.split('/', maxsplit=1)
        except ValueError:
            origin = 'foreground'
        return self._archive.catalog_ref(origin, external_ref)

    def count(self, entity_type):
        return self._archive.count_by_type(entity_type)

    def flows(self, **kwargs):
        for f in self._archive.search('flow', **kwargs):
            yield f

    def get_canonical(self, quantity):
        """
        By convention, a foreground archive's Term Manager is the catalog's LCIA engine, which is the Qdb of record
        for the foreground.
        :param quantity:
        :return:
        """
        return self._archive.tm.get_canonical(quantity)

    '''
    fg implementation begins here
    '''
    def fragments(self, *args, show_all=False, **kwargs):
        if hasattr(self._archive, 'fragments'):
            # we only want reference fragments by default
            for f in self._archive.fragments(*args, show_all=show_all, **kwargs):
                yield f
        else:
            raise NotForeground('The resource does not contain fragments: %s' % self._archive.ref)

    def context(self, item):
        return self._archive.tm[item]

    '''
    def frag(self, string, **kwargs):
        """
        :param string:
        :param kwargs: many=False
        :return:
        """
        return self._archive.frag(string, **kwargs)
    '''
    '''
    Create and modify fragments
    '''
    def new_quantity(self, name, ref_unit=None, **kwargs):
        """

        :param name:
        :param ref_unit:
        :param kwargs:
        :return:
        """
        q = new_quantity(name, ref_unit, **kwargs)
        self._archive.add(q)
        return q

    def add_or_retrieve(self, external_ref, reference, name, group=None, strict=False, **kwargs):
        """
        Gets a flow with the given external_ref, and creates it if it doesn't exist from the given spec.

        Note that the full spec is mandatory so it could be done by tuple.  With strict=False, an entity will be
        returned if it exists.
        :param external_ref:
        :param reference: a string, either a unit or known quantity
        :param name: the object's name
        :param group: used as context for flow
        :param strict: [False] if True it will raise a TypeError (ref) or ValueError (name) if an entity exists but
        does not match the spec.  n.b. I think I should check the reference entity even if strict is False but.. nah
        :return:
        """
        try:
            t = self.get(external_ref)
            if strict:
                if t.entity_type == 'flow':
                    if t.reference_entity != self.get_canonical(reference):
                        raise TypeError("ref quantity (%s) doesn't match supplied (%s)" % (t.reference_entity, reference))
                elif t.entity_type == 'quantity':
                    if t.unit != reference:
                        raise TypeError("ref unit (%s) doesn't match supplied (%s)" % (t.unit, reference))
                if t['Name'] != name:
                    raise ValueError("Name (%s) doesn't match supplied(%s)" % (t['Name'], name))
            return t

        except EntityNotFound:
            try:
                return self.new_flow(name, ref_quantity=reference, external_ref=external_ref, context=group, **kwargs)
            except UnknownRefQuantity:
                # assume reference is a unit string specification
                return self.new_quantity(name, ref_unit=reference, external_ref=external_ref, group=group, **kwargs)

    def new_flow(self, name, ref_quantity=None, context=None, **kwargs):
        """

        :param name:
        :param ref_quantity: defaults to "Number of items"
        :param context: [None] pending context refactor
        :param kwargs:
        :return:
        """
        if CONTEXT_STATUS_ == 'compat':
            if context is not None and 'compartment' not in kwargs:
                kwargs['compartment'] = str(context)
        if ref_quantity is None:
            ref_quantity = 'Number of items'
        try:
            ref_q = self.get_canonical(ref_quantity)
        except EntityNotFound:
            raise UnknownRefQuantity(ref_quantity)
        f = new_flow(name, ref_q, **kwargs)
        self._archive.add_entity_and_children(f)
        return self.get(f.link)

    def find_term(self, term_ref, origin=None, **kwargs):
        """

        :param term_ref:
        :param origin:
        :param kwargs:
        :return:
        """
        if term_ref is None:
            return
        if hasattr(term_ref, 'entity_type'):
            found_ref = term_ref
        else:
            # first context
            cx = self._archive.tm[term_ref]
            if cx is not None:
                found_ref = cx
            else:
                found_ref = self.get_local('/'.join(filter(None, (origin, term_ref))))
                ''' # this is now internal to get()
                except EntityNotFound:
                    if origin is None:
                        try:
                            origin, external_ref = term_ref.split('/', maxsplit=1)
                        except ValueError:
                            origin = 'foreground'
                            external_ref = term_ref

                        found_ref = self._archive.catalog_ref(origin, external_ref)
                    else:
                        found_ref = self._archive.catalog_ref(origin, term_ref)
                '''

        if found_ref.entity_type in ('flow', 'process', 'fragment', 'context'):
            return found_ref
        raise TypeError('Invalid entity type for termination: %s' % found_ref.entity_type)

    def new_fragment(self, flow, direction, external_ref=None, **kwargs):
        """
        :param flow:
        :param direction:
        :param external_ref: if provided, observe and name the fragment after creation
        :param kwargs: uuid=None, parent=None, comment=None, value=None, units=None, balance=False;
          **kwargs passed to LcFragment
        :return:
        """
        if isinstance(flow, str):
            try:
                f = self.get(flow)
            except EntityNotFound:
                f = self.get_local(flow)
            if f.entity_type != 'flow':
                raise TypeError('%s is not a flow' % flow)
            flow = f
        frag = create_fragment(flow, direction, origin=self.origin, **kwargs)
        self._archive.add_entity_and_children(frag)
        if external_ref is not None:
            self.observe(frag, name=external_ref)
        return frag

    '''
    This is officially deprecated. let it break.
    def name_fragment(self, fragment, name, auto=None, force=None, **kwargs):
        return self._archive.name_fragment(fragment, name, auto=auto, force=force)
    '''

    def observe(self, fragment, exchange_value=None, name=None, scenario=None, units=None, auto=None, force=None,
                accept_all=None, **kwargs):
        if accept_all is not None:
            print('%s: cannot "accept all"' % fragment)
        if name is not None and scenario is None:  #
            if fragment.external_ref != name:
                print('Naming fragment %s -> %s' % (fragment.external_ref, name))
                self._archive.name_fragment(fragment, name, auto=auto, force=force)
        if fragment.observable(scenario):
            if fragment not in self._observations:
                self._observations.append(fragment)
            if exchange_value is None:
                exchange_value = fragment.cached_ev  # do not expose accept_all via the interface; instead allow implicit
            fragment.observe(scenario=scenario, value=exchange_value, units=units)
        else:
            if exchange_value is not None:
                print('Note: Ignoring exchange value %g for unobservable fragment %s [%s]' % (exchange_value,
                                                                                              fragment.external_ref,
                                                                                              scenario))
        return fragment.link

    @property
    def observed_flows(self):
        for k in self._observations:
            yield k

    def fragments_with_flow(self, flow, direction=None, reference=None, background=None, **kwargs):
        """
        Requires flow identity
        :param flow:
        :param direction:
        :param reference:
        :param background:
        :param kwargs:
        :return:
        """
        flow = self[flow]  # retrieve by external ref
        for f in self._archive.fragments_with_flow(flow):
            if background is not None:
                if f.is_background != background:
                    continue
            if direction is not None:
                if f.direction != direction:
                    continue
            if reference is False and f.parent is None:
                continue
            if reference and f.parent:
                continue
            yield f

    def clone_fragment(self, frag, **kwargs):
        """

        :param frag: the fragment (and subfragments) to clone
        :param kwargs: suffix (default: ' (copy)', applied to Name of top-level fragment only)
                       comment (override existing Comment if present; applied to all)
        :return:
        """
        clone = clone_fragment(frag, **kwargs)
        self._archive.add_entity_and_children(clone)
        return clone

    def split_subfragment(self, fragment, replacement=None, descend=False, **kwargs):
        """
        Given a non-reference fragment, split it off into a new reference fragment, and create a surrogate child
        that terminates to it.

        without replacement:
        Old:   ...parent-->fragment
        New:   ...parent-->surrogate#fragment;   (fragment)

        with replacement:
        Old:   ...parent-->fragment;  (replacement)
        New:   ...parent-->surrogate#replacement;  (fragment);  (replacement)

        :param fragment:
        :param replacement: [None] if non-None, the surrogate is terminated to the replacement instead of the fork.
        :param descend: [False] on new term
        :return:
        """
        if fragment.reference_entity is None:
            raise AttributeError('Fragment is already a reference fragment')
        if replacement is not None:
            if replacement.reference_entity is not None:
                raise InvalidParentChild('Replacement is not a reference fragment')

        surrogate = _fork_fragment(fragment, comment='New subfragment')
        self._archive.add_entity_and_children(surrogate)

        fragment.unset_parent()
        if replacement is None:
            surrogate.terminate(fragment, descend=descend)
        else:
            surrogate.terminate(replacement, descend=descend)

        return fragment

    def delete_fragment(self, fragment, **kwargs):
        """
        Remove the fragment and all its subfragments from the archive (they remain in memory)
        This does absolutely no safety checking.
        :param fragment:
        :return:
        """
        if isinstance(fragment, str):
            try:
                fragment = self.get(fragment)
            except EntityNotFound:
                return False
        self._archive.delete_fragment(fragment)
        for c in fragment.child_flows:
            self.delete_fragment(c)
        return True

    def save(self, **kwargs):
        self._archive.save(**kwargs)
        return True

    def traverse(self, fragment, scenario=None, **kwargs):
        frag = self._archive.retrieve_or_fetch_entity(fragment)
        return frag.traverse(scenario, observed=True)

    def clear_unit_scores(self, lcia_method=None):
        self._archive.clear_unit_scores(lcia_method)

    def clear_scenarios(self, terminations=True):
        for f in self._archive.entities_by_type('fragment'):
            f.clear_scenarios(terminations=terminations)

    def create_process_model(self, process, ref_flow=None, set_background=True, **kwargs):
        rx = process.reference(ref_flow)
        if process.reference_value(ref_flow) < 0:  # put in to handle Ecoinvent treatment processes
            dirn = comp_dir(rx.direction)
        else:
            dirn = rx.direction
        frag = self.new_fragment(rx.flow, dirn, value=1.0, **kwargs)
        frag.terminate(process, term_flow=rx.flow)
        if set_background:
            frag.set_background()
        # self.fragment_from_exchanges(process.inventory(rx), parent=frag,
        #                              include_context=include_context, multi_flow=multi_flow)
        return frag

    def extend_process(self, fragment, scenario=None, **kwargs):
        term = fragment.termination(scenario)
        if not term.is_process:
            raise TypeError('Termination is not to process')
        fragment.unset_background()
        process = term.term_node
        self.fragment_from_exchanges(process.inventory(ref_flow=term.term_flow), parent=fragment, scenario=scenario,
                                     **kwargs)



    '''
    def extend_process_model(self, fragment, include_elementary=False, terminate=True, **kwargs):
        """
        "Build out" a fragment, creating child flows from its terminal node's intermediate flows
        :param fragment:
        :param include_elementary:
        :param terminate:
        :param kwargs:
        :return:
        """
        fragment.unset_background()
        process = fragment.term.term_node
        rx = fragment.term.term_flow
        for ex in process.inventory(rx):
            if not include_elementary:
                if ex.type in ('context', 'elementary'):
                    continue
            ch = self.new_fragment(ex.flow, ex.direction, value=ex.value, parent=fragment)
            ch.set_background()
            if ex.type in ('cutoff', 'self'):
                continue
            if terminate:
                ch.terminate(self._archive.catalog_ref(process.origin, ex.termination, entity_type='process'),
                             term_flow=ex.flow)
        fragment.observe(accept_all=True)
        return fragment
    '''

    '''# Create or update a fragment from a list of exchanges.

    This needs to be an interface method.
    '''
    def fragment_from_exchanges(self, _xg, parent=None, ref=None, scenario=None,
                                set_background=True,
                                include_context=False,
                                multi_flow=False):
        """
        If parent is None, first generated exchange is reference flow; and subsequent exchanges are children.
        Else, all generated exchanges are children of the given parent, and if a child flow exists, update it.

        We can take two approaches: in the simple view, the child flow is unique, and the termination can be updated
        by a subsequent specification.  This obviously fails if a fragment has multiple children with the same flow--
        as is in fact the case with ecoinvent.  So for that case we just need to match on termination as well; but
        that means a termination cannot get updated automatically through this process.

        This is all tricky if we expect it to work with both ExchangeRefs and actual exchanges (which, obviously, we
        should) because: ExchangeRefs may have only a string for process and flow, but Exchanges will have entities
        for each.  We need to get(flow_ref) and find_term(term_ref) but we can simply use flow entities and catalog
        refs if we have process.origin.  For now we will just swiss-army-knife it.

        :param _xg: Generates a list of exchanges or exchange references
        :param parent: if None, create parent from first exchange
        :param ref: if parent is created, assign it a name (if parent is non-None, ref is ignored
        :param set_background: [True] whether to regard process-terminated fragments as background fragments
        :param include_context: [False] whether to model context-terminated flows as child fragments
        :param multi_flow: [False] if True, child flows are matched on flow, direction, and termination
        :return:
        """
        if parent is None:
            x = next(_xg)
            parent = self.new_fragment(x.flow, x.direction, value=x.value, units=x.unit, Name=str(x.process), **x.args)
            if ref is None:
                print('Creating new fragment %s (%s)' % (x.process.name, parent.uuid))
            else:
                print('Creating new fragment %s' % ref)
                self.observe(parent, name=ref)
            update = False
        else:
            update = True

        for y in _xg:
            if hasattr(y.flow, 'entity_type') and y.flow.entity_type == 'flow':
                flow = y.flow
            else:
                flow = self[y.flow]
                if flow is None:
                    print('Skipping unknown flow %s' % y.flow)
                    continue
            if hasattr(y.process, 'origin'):
                term = self.find_term(y.termination, origin=y.process.origin)
            else:
                term = self.find_term(y.termination)
            if term is not None and term.entity_type == 'context' and include_context is False:
                continue
            if term == y.process:
                term = None  # don't terminate self-term
            if update:
                try:
                    # TODO: children_with_flow needs to be scenario aware
                    if multi_flow:
                        c_up = next(parent.children_with_flow(flow, direction=y.direction, termination=term,
                                                              recurse=False))
                    else:
                        c_up = next(parent.children_with_flow(flow, direction=y.direction, recurse=False))

                    # update value
                    v = y.value
                    if y.unit is not None:
                        v *= c_up.flow.reference_entity.convert(y.unit)

                    if c_up.exchange_value(scenario) != v:
                        print('Updating %s exchange value %.3f' % (c_up, v))

                        self.observe(c_up, exchange_value=v, scenario=scenario)
                    if multi_flow:
                        continue  # cannot update terms in the multi-flow case

                    # set term
                    if term is not None:
                        if term != c_up.term.term_node:
                            print('Updating %s termination %s' % (c_up, term))
                            c_up.clear_termination(scenario)
                            if term.entity_type == 'fragment':
                                c_up.terminate(term, scenario=scenario)
                            if term.entity_type == 'process' and set_background:
                                c_up.set_background()
                    continue
                except StopIteration:
                    print('No child flow found; creating new %s %s' % (flow, y.direction))
                    pass

            c = self.new_fragment(flow, y.direction, value=y.value, units=y.unit, parent=parent, **y.args)

            if term is not None:
                c.terminate(term, scenario=scenario)
                if term.entity_type == 'process' and set_background:
                    c.set_background()
            self.observe(c)  # use cached implicitly via fg interface

        return parent

    def make_fragment_trees(self, exchanges):
        """
        Take in a list of exchanges [that are properly connected] and build fragment trees from them. Return all roots.

        If an exchange's process has been encountered, it will be used as the parent.  Exchanges with null terminations
        become cutoff flows.

        If an exchange's process has not been encountered, AND its termination is null, it will become a reference
        fragment and be terminated to the process.

        Non-null-terminated exchanges whose processes have not been encountered cause an error.

        This function will only generate new fragments and will not affect any existing fragments.
        :param exchanges:
        :return:
        """
        roots = []
        parents = dict()
        for x in exchanges:
            if x.process.external_ref in parents:
                parent = parents[x.process.external_ref]
                parent.unset_background()
                frag = self.new_fragment(x.flow, x.direction, parent=parent, **x.args)
                term = self.find_term(x.termination, origin=x.process.origin)
                if term is not None:
                    frag.terminate(term)
                    frag.set_background()
            else:
                # unmatched flows become roots
                if x.termination is not None:
                    raise InvalidParentChild('Reference flow may not be terminated')
                frag = self.new_fragment(x.flow, x.direction, **x.args)
                roots.append(frag)
                frag.terminate(x.process)

            self.observe(frag, exchange_value=x.value)
            if frag.term.is_process:
                parents[frag.term.term_ref] = frag

        for r in roots:
            yield r



from antelope import local_ref
from antelope_core.catalog import LcCatalog

from shutil import rmtree
import os
import re


class ForegroundCatalog(LcCatalog):
    """
    Adds the ability to create (and manage?) foreground resources
    """

    '''
    ForegroundCatalog
    '''

    def foreground(self, ref, path=None, quiet=True, reset=False, delete=False):
        """
        Creates or activates a foreground resource and returns an interface to that resource.
        By default creates in a subdirectory of the catalog root with the ref as the folder
        :param ref:
        :param path:
        :param quiet:
        :param reset:
        :param delete:
        :return:
        """
        """
        Creates or activates a foreground as a sub-folder within the catalog's root directory.  Returns a
        Foreground interface.
        :param path: either an absolute path or a subdirectory path relative to the catalog root
        :param ref: semantic reference (optional)
        :param quiet: passed to fg archive
        :param reset: [False] if True, clear the archive and create it from scratch, before returning the interface
        :param delete: [False] if True, delete the existing tree completely and irreversibly. actually just rename
        the directory to whatever-DELETED; but if this gets overwritten, there's no going back.  Overrides reset.
        :return:
        """
        if path is None:
            path = os.path.join(self._rootdir, ref)  # should really sanitize this somehow
            localpath = ref
        else:
            if os.path.isabs(path):
                localpath = None
            else:
                localpath = path
                path = os.path.join(self._rootdir, path)


        abs_path = os.path.abspath(path)
        local_path = self._localize_source(abs_path)

        if delete:
            if localpath:
                if os.path.exists(localpath):
                    del_path = localpath + '-DELETED'
                    if os.path.exists(del_path):
                        rmtree(del_path)
                    os.rename(abs_path, del_path)
            dels = [k for k in self._resolver.resolve(ref, interfaces='foreground')]
            for k in dels:
                self.delete_resource(k, delete_source=True, delete_cache=True)

        try:
            res = next(self._resolver.resources_with_source(local_path))
        except StopIteration:
            res = self.new_resource(ref, local_path, 'LcForeground', interfaces=['index', 'foreground', 'quantity'],
                                    quiet=quiet)

        if reset:
            res.remove_archive()
        res.check(self)

        return res.make_interface('foreground')

    def assign_new_ref(self, old_ref, new_ref):
        """
        This only works for certain types of archives. Foregrounds, in particular. but it is hard to say what else.
        What needs to happen here is:
         - first we retrieve the archive for the ref (ALL archives?)
         - then we call set_origin() on the archive
         - then we save the archive
         - then we rename the resource file
         = actually we just rewrite the resource file, since the filename and JSON key have to match
         = since we can't update resource references, it's easiest to just blow them away and reload them
         = but to save time we should transfer the archives from the old resource to the new resource
         = anyway, it's not clear when we would want to enable this operation in the first place.
         * so for now we leave it
        :param old_ref:
        :param new_ref:
        :return:
        """
        pass

    def configure_resource(self, reference, config, *args):
        """
        We must propagate configurations to internal, derived resources. This also begs for testing.
        :param reference:
        :param config:
        :param args:
        :return:
        """
        # TODO: testing??
        for res in self._resolver.resolve(reference, strict=False):
            abs_src = self.abs_path(res.source)
            if res.add_config(config, *args):
                if res.internal:
                    if os.path.dirname(abs_src) == self._index_dir:
                        print('Saving updated index %s' % abs_src)
                        res.archive.write_to_file(abs_src, gzip=True,
                                                  exchanges=False, characterizations=False, values=False)
                else:
                    print('Saving resource configuration for %s' % res.reference)
                    res.save(self)

            else:
                if res.internal:
                    print('Deleting unconfigurable internal resource for %s\nsource: %s' % (res.reference, abs_src))
                    self.delete_resource(res, delete_source=True)
                else:
                    print('Unable to apply configuration to resource for %s\nsource: %s' % (res.reference, res.source))



"""
Flow Terminations are model-defined links between a particular flow and a process that terminates it.

They originated as part of LcFragments but in fact they are more general. A FlowTermination is actually the same
as a ProductFlow in lca-matrix, plus features to compute LCIA.  It should be easy to construct either one from the
other.
"""

from antelope import PrivateArchive, check_direction, comp_dir, NoFactorsFound, QuantityRequired, MultipleReferences

from antelope_core.exchanges import ExchangeValue
from antelope_core.lcia_results import LciaResult
from .lcia_dict import LciaResults


# from lcatools.catalog_ref import NoCatalog
# from lcatools.interact import parse_math


class FlowConversionError(Exception):
    pass


class SubFragmentAggregation(Exception):
    pass


class NonConfigurableInboundEV(Exception):
    """
    only foreground terminations may have their inbound exchange values explicitly specified
    """
    pass


class UnCachedScore(Exception):
    """
    means that we have an LCIA-only node whose score has not been set for the requested LCIA method
    """
    pass


class TerminationFromJson(Exception):
    """
    Something failed in deserialization
    """
    pass


class FlowTermination(object):

    _term = None
    _term_flow = None
    _direction = None
    _descend = True

    """
    these are stored by scenario in a dict on the mainland

    A fragment can have the following types of terminations:
     * None - the termination is null- the flow enters the foreground and becomes an i/o
     * parent - the fragment's termination is the fragment itself.  The fragment flow  enters a foreground node.
       The node can have children but only has LCIA impacts based on the terminating flow, which have to be looked up
       in the database. fg-terminated nodes don't have scenarios (e.g. the scenarios are in the exchange values).
       Note: term_flows can be different from parent flows, and unit conversions will occur normally (e.g. "sulfur
       content" converted to "kg SO2")
     * Process - the flow enters a process referenced by CatalogRef.  The node's LCIA impacts are fg_lcia. The
       node's children are the process's non-term intermediate exchanges. The node can also have other children.
       (created with terminate or term_from_exch)
     * Fragment - the flow enters a sub-fragment.  The sub-fragment must be traversable. The node's children are
       the fragment's non-term io flows. The node cannot have other children.  If the sub-fragment is background,
       then the background fragment flow supplants the foreground one during traversal.

    LCIA results are always cached in the terminations, and are not (by default) persistent across instantiations.
    """
    @classmethod
    def from_json(cls, fragment, fg, scenario, j):
        if len(j) == 0:
            return cls.null(fragment)
        origin = j.pop('source', None) or j.pop('origin')
        if origin == 'foreground':
            origin = fg.ref

        # handle term flow
        tf_ref = j.pop('termFlow', None)
        if tf_ref is None:
            term_flow = None  # if it is unspecified, let the best value be determined
        elif isinstance(tf_ref, dict):
            term_flow = fg.catalog_ref(tf_ref['origin'], tf_ref['externalId'], entity_type='flow')
        else:
            if origin == fg.ref:
                term_flow = fg[tf_ref]
            else:
                term_flow = fg.catalog_ref(origin, tf_ref, entity_type='flow')

        if 'context' in j:
            term_node = fg.tm[j['context']]
        else:
            try:
                external_ref = j['externalId']
            except KeyError:
                external_ref = j['entityId']
            # handle term_node
            if origin == fg.ref:
                term_node = fg[external_ref]
            else:
                term_node = fg.catalog_ref(origin, external_ref, entity_type='process')

        direction = j.pop('direction', None)
        descend = j.pop('descend', True)
        try:
            term = cls(fragment, term_node, _direction=direction, term_flow=term_flow, descend=descend)
        except MultipleReferences:
            raise TerminationFromJson('term_flow missing and ambiguous: %s (%s)' % (fragment.link, scenario))
        if 'scoreCache' in j.keys():
            term._deserialize_score_cache(fg, j['scoreCache'], scenario)
        return term

    '''
    @classmethod
    def from_exchange_ref(cls, fragment, exchange_ref):
        return cls(fragment, exchange_ref.process_ref, direction=exchange_ref.direction,
                   term_flow=exchange_ref.exchange.flow, inbound_ev=exchange_ref.exchange.value)

    @classmethod
    def from_term(cls, fragment, term):
        return cls(fragment, term.term_node, direction=term.direction, term_flow=term.term_flow,
                   descend=term.descend, inbound_ev=term.inbound_exchange_value)
    '''

    @classmethod
    def null(cls, fragment):
        return cls(fragment, None)

    def __init__(self, fragment, entity, _direction=None, term_flow=None, descend=True, inbound_ev=None):
        """
        reference can be None, an entity or a catalog_ref.  It only must have origin, external_ref, and entity_type.
        To use an exchange, use FlowTermination.from_exchange()
         * None - to create a foreground IO / cutoff flow
         * fragment (same as parent) - to create a foreground node.  Must satisfy 'fragment is entity'
         * process or process ref - to link the fragment to a process inventory (uses self.is_background to determine
           foreground or background lookup)
         * context - to represent the fragment's flow (or term_flow, still supported) as emission
         * flow or flow ref - no longer supported.  Supply context instead.

        The term's direction is detected at the time of creation.

        :param fragment:
        :param entity:
        :param _direction: Should not be used externally- only if term is generated from_json
        :param term_flow: optional flow to match on termination inventory or LCIA.  If flow and term_flow have
        different reference quantities, quantity conversion is performed during traversal
        :param descend:
        :param inbound_ev: ignored; deprecated
        """

        self._parent = fragment
        if entity is not None:
            if entity.entity_type == 'flow':
                if term_flow is not None and term_flow != entity:
                    raise ValueError('Inconsistent flow and term_flow provided: %s' % entity)
                term_flow = entity
                entity = fragment  # foreground termination with flow conversion
                # raise TypeError('Can no longer terminate fragments with flows. Use context instead')
            elif entity.entity_type not in ('context', 'process', 'fragment'):
                raise TypeError('Inappropriate termination type: %s' % entity.entity_type)
        self._term = entity  # this must have origin, external_ref, and entity_type, and be operable (if ref)
        self._score_cache = LciaResults(fragment)

        self.term_flow = term_flow
        self.direction = _direction
        self.descend = descend

    @property
    def term_flow(self):
        if self._term_flow is None:
            if self.is_process:
                raise AttributeError('[%s] term_flow was not specified for process term!' % self._parent.external_ref)
            if self.is_frag:
                return self.term_node.flow
            else:
                return self._parent.flow
        return self._term_flow

    @term_flow.setter
    def term_flow(self, term_flow):
        """
        Introduce term validation checking here if needed
        :param term_flow:
        :return:
        """
        if term_flow is None:
            if self.is_process:
                self._term_flow = self._term.reference().flow
            elif self.is_frag:
                self._term_flow = None  # leave unspecified to plug into term's ref flow
            else:
                self._term_flow = self._parent.flow
        else:
            # TODO: check to see if supplied term flow is valid / can be a reference flow for term
            self._term_flow = term_flow
        if self.valid and self.node_weight_multiplier == 0:
            print('Warning: 0 node weight multiplier for term of %s' % self._parent.external_ref)

    @property
    def direction(self):
        return self._direction

    @property
    def valid(self):
        if self.is_null:
            return False
        return self.term_node.validate()

    @direction.setter
    def direction(self, value):
        if value is None:
            # this is the default: should set the direction by the reference.  Only non-none if from_json
            if self.is_process and self.valid:
                rx = self.term_node.reference(self.term_flow)
                value = rx.direction
            else:
                # for fg, invert direction doesn't make sense. for subfragments, direction is ignored
                value = comp_dir(self._parent.direction)
        self._direction = check_direction(value)

    '''
    def matches(self, exchange):
        """
        returns True if the exchange specifies the same process and flow as the term's process_ref and term_flow
        :param exchange:
        :return:
        """
        if self.is_null:
            return False
        if self.term_node.entity_type != 'process':
            return False
        return (self._term.external_ref == exchange.process.external_ref) and (self.term_flow.match(exchange.flow))

    def terminates(self, exchange):
        """
        Returns True if the exchange's termination matches the term's term_node, and the flows also match, and the
        directions are complementary.
        If the exchange does not specify a termination, returns True if the flows match and directions are comp.
        :param exchange:
        :return:
        """
        if self.term_flow.match(exchange.flow) and self.direction == comp_dir(exchange.direction):
            if exchange.termination is None:
                return True
            else:
                if self.is_null:
                    return False
                if self.term_node.entity_type != 'process':
                    return False
                if exchange.termination == self._term.external_ref:
                    return True
        return False

    def to_exchange(self):
        if self.is_null:
            return None
        return ExchangeValue(self.term_node, self.term_flow, self.direction, value=self.inbound_exchange_value)
    '''

    @property
    def is_local(self):
        """
        Fragment and termination have the same origin
        :return:
        """
        if self.is_null:
            return False
        return self._parent.origin == self.term_node.origin

    @property
    def is_context(self):
        """
        termination is a context
        :return:
        """
        if self.is_null:
            return False
        return self.term_node.entity_type == 'context'

    @property
    def is_frag(self):
        """
        Termination is a fragment
        :return:
        """
        return (not self.is_null) and (self.term_node.entity_type == 'fragment')

    @property
    def is_process(self):
        """
        termination is a process
        :return:
        """
        return (not self.is_null) and (self.term_node.entity_type == 'process')

    @property
    def is_emission(self):
        """
        Pending context refactor
        :return:
        """
        return self.is_context and self.term_node.elementary

    @property
    def is_fg(self):
        """
        Termination is parent
        :return:
        """
        return (not self.is_null) and (self.term_node is self._parent)

    @property
    def is_bg(self):
        """
        parent is marked background, or termination is a background fragment
        :return:
        """
        return self._parent.is_background

    @property
    def term_is_bg(self):
        """
        Termination is local and background
        :return:
        """
        return self.is_frag and self.is_local and self.term_node.is_background

    @property
    def is_subfrag(self):
        """
        Termination is a non-background, non-self fragment.
        Controversy around whether expression should be:
        self.is_frag and not (self.is_fg or self.is_bg or self.term_is_bg)  [current] or
        self.is_frag and (not self.is_fg) and (not self.is_bg)  [old; seems wrong]

        :return:
        """
        return self.is_frag and not (self.is_fg or self.is_bg or self.term_is_bg)

    @property
    def is_null(self):
        return self._term is None

    @property
    def descend(self):
        return self._descend

    @descend.setter
    def descend(self, value):
        if value is None:
            return
        if isinstance(value, bool):
            self._descend = value
            ''' # this whole section not needed- we can certainly cache LCIA scores for nondescend fragments,
            and we don't need to blow them away if descend is True; just ignore them.
            if value is True:
                self.clear_score_cache()  # if it's descend, it should have no score_cache
                # if it's not descend, the score gets computed (and not cached) during traversal
            '''
        else:
            raise ValueError('Descend setting must be True or False')

    @property
    def term_node(self):
        return self._term

    @property
    def term_ref(self):
        if self.is_null:
            return None
        elif self.is_context:
            return self.term_node.name
        return self.term_node.external_ref

    @property
    def flow_conversion(self):
        """
        express the parent's flow in terms of the quantity of the term flow
        how to deal with scenario cfs? tbd
        problem is, the term doesn't know its own scenario
        :return: float = amount in term_flow ref qty that corresponds to a unit of fragment flow's ref qty
        """
        if not self.valid:
            return 0.0
        if self.term_flow.reference_entity == self._parent.flow.reference_entity:
            return 1.0
        try:
            fwd_cf = self.term_flow.reference_entity.cf(self._parent.flow, dist=3)
        except QuantityRequired:
            fwd_cf = 0.0
        if fwd_cf == 0.0:
            try:
                rev_cf = self._parent.flow.reference_entity.cf(self.term_flow, dist=3)
            except QuantityRequired:
                rev_cf = 0.0
            if rev_cf == 0.0:
                raise FlowConversionError('Zero CF found relating %s to %s' % (self.term_flow, self._parent.flow))
            else:
                return 1.0 / rev_cf
        else:
            return fwd_cf

    @property
    def id(self):
        if self.is_null:
            return None
        else:
            return self._term.external_ref

    @property
    def inbound_exchange_value(self):
        """
        This is only used for correcting fragment-term direction mismatches.
        This needs to be tested!
        :return:
        """
        if self.direction == self._parent.direction:
            return -1.0
        return 1.0

    @inbound_exchange_value.setter
    def inbound_exchange_value(self, val):
        raise NonConfigurableInboundEV

    @property
    def node_weight_multiplier(self):
        return self.flow_conversion / self.inbound_exchange_value

    @property
    def unit(self):
        if self.is_null:
            return '--'
        if self.term_node.entity_type == 'fragment':  # fg, bg, or subfragment
            return '%4g unit' % self.inbound_exchange_value
        return '%4g %s' % (self.inbound_exchange_value, self.term_flow.unit)  # process

    def _unobserved_exchanges(self):
        """
        Generator which yields exchanges from the term node's inventory that are not found among the child flows, for
          LCIA purposes

        Challenge here going forward: we made some kind of normative decision early on that terminations do not know
        their own scenarios, that the fragment maps scenario to termination. The problem is that now terminations
        cannot themselves carry out traversal on the term_node because they don't know what scenarios to pass.

        The upshot is that we cannot meaningfully compute "unobserved exchanges" for subfragments, since we don't
        know our scenario.

        :return:
        """
        if self.is_context:
            x = ExchangeValue(self._parent, self.term_flow, self._parent.direction, termination=self.term_node,
                              value=self.node_weight_multiplier)
            yield x
        elif self.is_frag:  # fragments can have unobserved exchanges too! (CAN THEY?)
            for x in []:
                yield x
        else:
            children = set()
            children.add((self.term_flow.external_ref, self.direction, None))
            for c in self._parent.child_flows:
                children.add((c.flow.external_ref, c.direction))
            if self.is_bg:
                iterable = self.term_node.lci(self.term_flow)
            else:
                iterable = self.term_node.inventory(ref_flow=self.term_flow)
            for x in iterable:
                if (x.flow.external_ref, x.direction) not in children:
                    yield x

    def compute_unit_score(self, quantity_ref, **kwargs):
        """
        four different ways to do this.
        0- we are a subfragment-- throw exception: use subfragment traversal results contained in the FragmentFlow
        1- parent is bg: ask catalog to give us bg_lcia (process or fragment)
        2- get fg lcia for unobserved exchanges

        If
        :param quantity_ref:
        :return:
        """
        if self.is_frag:
            '''
            if self.is_subfrag:
                if not self.descend:
                    raise SubFragmentAggregation  # to be caught

            #

            # either is_fg (no impact) or is_bg or term_is_bg (both equiv)

            elif self.is_bg:
                # need bg_lcia method for FragmentRefs
                # this is probably not currently supported
                # return self.term_node.bg_lcia(lcia_qty=quantity_ref, ref_flow=self.term_flow.external_ref, **kwargs)
                # instead- just do fragment_lcia
                print('Warning: ignoring spurious background setting for subfrag:\n%s\n%s' % (self._parent, self.term_node))
                return LciaResult(quantity_ref)

            else:
                assert self.is_fg

                # in the current pre-ContextRefactor world, this is how we are handling
                # cached-LCIA-score nodes
                # in the post-Context-Refactor world, foreground frags have no impact
                #raise UnCachedScore('fragment: %s\nquantity: %s' % (self._parent, quantity_ref))
                return LciaResult(quantity_ref)
            '''
            return LciaResult(quantity_ref)

        try:
            if self.is_context:
                locale = self._parent['SpatialScope']
            else:
                locale = self.term_node['SpatialScope']
        except KeyError:
            locale = 'GLO'
        try:
            res = quantity_ref.do_lcia(self._unobserved_exchanges(), locale=locale, **kwargs)
        except PrivateArchive:
            if self.is_bg:
                print('terminations.compute_unit_score UNTESTED for private bg archives!')
                res = self.term_node.bg_lcia(lcia_qty=quantity_ref, ref_flow=self.term_flow.external_ref, **kwargs)
            else:
                res = self.term_node.fg_lcia(quantity_ref, ref_flow=self.term_flow.external_ref, **kwargs)
                print('terminations.compute_unit_score UNTESTED for private fg archives!')

        res.scale_result(self.inbound_exchange_value)
        return res

    def score_cache(self, quantity=None, ignore_uncached=False, refresh=False, **kwargs):
        """
        only process-terminations are cached

        :param quantity:
        :param ignore_uncached:
        :param refresh: If True, re-compute unit score even if it is already present in the cache. This fails on
        multi-instance fragments by causing the
        :param kwargs:
        :return:
        """
        if quantity is None:
            return self._score_cache

        if self.is_frag:
            if self.is_subfrag:
                if not self.descend:
                    raise SubFragmentAggregation  # to be caught- subfrag needs to be queried w/scenario
            return LciaResult(quantity)  # otherwise, subfragment terminations have no impacts

        if quantity in self._score_cache and refresh is False:
            return self._score_cache[quantity]
        else:
            try:
                res = self.compute_unit_score(quantity, **kwargs)
            except UnCachedScore:
                if ignore_uncached:
                    res = LciaResult(quantity)
                else:
                    raise
            self._score_cache[quantity] = res
            return res

    def score_cache_items(self):
        return self._score_cache.items()

    def lcia(self):
        for k, v in self.score_cache_items():
            print('%s' % v)

    def reset_score(self, lcia):
        self._score_cache.pop(lcia, None)

    def clear_score_cache(self):
        self._score_cache.clear()

    def _serialize_score_cache(self):
        """
        Score cache contains an LciaResults object, which works as a dict.
        serialization should preserve order, which prohibits using a simple dict
        :return: a list to be serialized directly
        """
        score_cache = []
        for q in self._score_cache.indices():
            res = self._score_cache[q]
            score_cache.append({'quantity': {'origin': res.quantity.origin,
                                             'externalId': res.quantity.external_ref},
                                'score': res.total()})
        return score_cache

    def add_lcia_score(self, quantity, score, scenario=None):
        res = LciaResult(quantity, scenario=scenario)
        res.add_summary(self._parent.uuid, self._parent, 1.0, score)
        self._score_cache.add(res)

    def _deserialize_score_cache(self, fg, sc, scenario):
        self._score_cache = LciaResults(self._parent)
        for i in sc:
            q = fg.catalog_ref(i['quantity']['origin'], i['quantity']['externalId'], entity_type='quantity')
            self.add_lcia_score(q, i['score'], scenario=scenario)

    def _term_flow_block(self):
        if self.term_flow.origin == self.term_node.origin:
            return self.term_flow.external_ref
        else:
            return {
                'origin': self.term_flow.origin,
                'externalId': self.term_flow.external_ref
            }

    def serialize(self, save_unit_scores=False):
        if self.is_null:
            return {}
        if self.is_context:
            j = {
                'origin': self._term_flow.origin,
                'context': self._term.name
            }
        else:
            j = {
                'origin': self._term.origin,
                'externalId': self._term.external_ref
            }
        # saving term_flow: for subfragments, we save it only it it's specified
        if self.is_frag:
            if self._term_flow is not None:
                j['termFlow'] = self._term_flow_block()
        elif self.term_flow != self._parent.flow:
            j['termFlow'] = self._term_flow_block()

        if self.direction != comp_dir(self._parent.direction):
            j['direction'] = self.direction
        if self._descend is False:
            j['descend'] = False
        if self._parent.is_background and save_unit_scores and len(self._score_cache) > 0:
            j['scoreCache'] = self._serialize_score_cache()
        return j

    def __eq__(self, other):
        """
        Terminations are equal if they are both null, both fg, or if term_node, term_flow, direction and descend match
        :param other:
        :return:
        """
        if self is other:
            return True
        if not isinstance(other, FlowTermination):
            return False
        if self.is_null:
            if other.is_null:
                return True
            return False
        if self.is_fg:
            if other.is_fg:
                return True
            return False
        return (self.term_node.external_ref == other.term_node.external_ref and
                self.term_flow == other.term_flow and
                self.direction == other.direction and
                self.descend == other.descend)

    def __str__(self):
        """

        :return:
          '---:' = fragment I/O
          '-O  ' = foreground node
          '-*  ' = process
          '-#  ' - sub-fragment (aggregate)
          '-#::' - sub-fragment (descend)
          '-B ' - terminated background
          '--C ' - cut-off background
        """
        if self.is_null:
            term = '---:'  # fragment IO
        elif self.is_fg:
            term = '-O  '
        elif self.is_context:
            if self.is_emission:
                term = '-== '
            else:
                # TODO: intermediate contexts don't present as cutoffs (because is_null is False)
                term = '-cx '
        elif self.term_node.entity_type == 'process':
            if self.is_bg:
                term = '-B* '
            else:
                term = '-*  '
        elif self.term_node.entity_type == 'fragment':
            if self.term_is_bg:
                # TODO: Broken! needs to be scenario-aware
                if self.term_node.term.is_null:
                    term = '--C '
                else:
                    term = '-B  '
            else:
                if self.descend:
                    term = '-#::'
                else:
                    term = '-#  '
        else:
            raise TypeError('I Do not understand this term for frag %.7s' % self._parent.uuid)
        return term




from ...implementations import AntelopeForegroundImplementation


class AntelopeV1ForegroundImplementation(AntelopeForegroundImplementation):
    """
    Overrides the default implementation to handle the AntelopeV1 case
    """

    '''
    Implementation Method Overrides
    Unimplemented overrides will simply fallthrough to default:

    def exchanges(self, process, **kwargs):
        pass

    def exchange_values(self, process, flow, direction, termination=None, **kwargs):
        pass

    def inventory(self, process, ref_flow=None, **kwargs):
        pass
    '''

    def traverse(self, fragment, scenario=None, **kwargs):
        if scenario is not None:
            endpoint = 'scenarios/%s/%s/fragmentflows' % (scenario, fragment)
        else:
            endpoint = '%s/fragmentflows' % fragment

        self._archive.fetch_flows(fragment)

        ffs = self._archive.get_endpoint(endpoint)
        for ff in ffs:
            if 'fragmentStageID' in ff:
                ff['StageName'] = self._archive.get_stage_name(ff['fragmentStageID'])
        return [self._archive.make_fragment_flow(ff)
                for ff in sorted(ffs, key=lambda x: ('parentFragmentFlowID' in x, x['fragmentFlowID']))]



from antelope.refs.base import EntityRef
# from ..fragment_flows import group_ios
"""
Not sure what to do about Fragment Refs, whether they belong in the main interface. I'd like to think no, but
for now we will just deprecate them and remove functionality,
"""


class FragmentRef(EntityRef):
    """
    Fragments can lookup:
    """
    '''
    def __init__(self, *args, **kwargs):
        super(FragmentRef, self).__init__(*args, **kwargs)
        self._known_scenarios = dict()
    '''
    _etype = 'fragment'
    _ref_field = 'parent'

    def __init__(self, *args, **kwargs):
        super(FragmentRef, self).__init__(*args, **kwargs)
        self._direction = None
        self._flow = None
        self._isset = False

    def set_config(self, flow, direction):
        if self._isset:
            raise AttributeError('Fragment Ref is already specified!')
        self._isset = True
        self._flow = flow
        self._direction = direction

    @property
    def direction(self):
        return self._direction

    @property
    def is_background(self):
        """
        Can't figure out whether it ever makes sense for a fragment ref to be regarded 'background'
        :return:
        """
        return False

    @property
    def flow(self):
        return self._flow

    @property
    def _addl(self):
        return 'frag'

    @property
    def name(self):
        if self.external_ref is None:
            return self['Name']
        return self.external_ref

    @property
    def is_conserved_parent(self):
        return None

    def set_name(self, name, **kwargs):
        return self._query.name_fragment(self, name, **kwargs)

    def inventory(self, scenario=None, **kwargs):
        return self._query.inventory(self.external_ref, scenario=scenario, **kwargs)

    def traverse(self, scenario=None, **kwargs):
        return self._query.traverse(self.external_ref, scenario=scenario, **kwargs)

    def fragment_lcia(self, lcia_qty, scenario=None, **kwargs):
        return self._query.fragment_lcia(self.external_ref, lcia_qty, scenario=scenario, **kwargs)

    def bg_lcia(self, lcia_qty, scenario=None, **kwargs):
        return self.fragment_lcia(self.external_ref, lcia_qty, scenario=scenario, **kwargs)

    def unit_inventory(self, scenario=None, observed=None):
        """

        :param scenario:
        :param observed: ignored; supplied only for signature consistency
        :return:
        """
        '''
        if observed is False:
            print('Ignoring false observed flag')
        ffs = self.traverse(scenario=scenario)  # in the future, may want to cache this
        return group_ios(self, ffs)
        '''
        return NotImplemented



from setuptools import setup, find_packages

requires = [
    'antelope_core'
]

"""
Revision history

0.1.3 - 30 Dec 2020 - First public release
"""

setup(
    name="antelope_foreground",
    version="0.1.3rc1",
    author="Brandon Kuczenski",
    author_email="bkuczenski@ucsb.edu",
    install_requires=requires,
    url="https://github.com/AntelopeLCA/foreground",
    summary="A foreground model building implementation",
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    packages=find_packages()
)



"""
A foreground archive is an LcArchive that also knows how to access, serialize, and deserialize fragments.
"""
import json
import os
import re

from collections import defaultdict

from ..implementations import AntelopeForegroundImplementation as ForegroundImplementation

from antelope import PropertyExists
from antelope_core.archives import BasicArchive, EntityExists, BASIC_ENTITY_TYPES
from ..entities.fragments import LcFragment


FOREGROUND_ENTITY_TYPES = BASIC_ENTITY_TYPES + ('fragment', )


class AmbiguousReference(Exception):
    pass


class FragmentNotFound(Exception):
    pass


class FragmentMismatch(Exception):
    pass


class NonLocalEntity(Exception):
    """
    Foregrounds should store only local entities and references to remote entities
    """
    pass


class LcForeground(BasicArchive):
    """
    An LcForeground is defined by being anchored to a physical directory, which is used to serialize the non-fragment
    entities.  Also within this directory is a subdirectory called fragments, which is used to store fragments
    individually as files.

    Fragments are "observations" of foreground systems. A flow is observed to pass from / into some reference entity
    and is terminated (other end of the flow) to x, which can either be None (a cutoff), or an entity ID which is
    dereferenced to a context (=> elementary) or a process (=>intermediate).

    A foreground is then a collection of observations of flows passing through processes.

    Foreground models can be constructed flow by flow (observed from unit process inventories
    """
    _entity_types = FOREGROUND_ENTITY_TYPES
    _ns_uuid_required = None

    def _load_entities_json(self, filename):
        with open(filename, 'r') as fp:
            self.load_from_dict(json.load(fp), jsonfile=filename)

    def _ref_to_key(self, key):
        """
        In a foreground, there are three different ways to retrieve an entity.

        (0) by link.  The master key for each entity is its link, which is its key in _entities.

        (1) by uuid.  fg archives have no nsuuid capabilities, but if an entity has a uuid, it can be retrieved by it.

        (2) by name.  Fragments and any other entities native to the foreground archive can also be retrieved by their
        external ref.  Fragments are the only entities that can be *assigned* a name after creation.  [for this reason,
        Fragments MUST have UUIDs to be used for hashing]

         * _ref_to_key transmits the user's heterogeneous input into a valid key to self._entities, or None
           - if the key is already a known link, it's easy
           - if the key is a custom name, it's almost as easy
           - if the key CONTAINS a uuid known to the db, then a link for an entity with that UUID is
              NONDETERMINISTICALLY returned
         * __getitem__ uses _ref_to_key to translate the users input into a key

        For the thorns around renaming fragments, see self.name_fragment() below

        :param key:
        :return:
        """
        if key in self._entities:
            return key
        elif key in self._ext_ref_mapping:
            return self._ext_ref_mapping[key]
        else:
            uid = self._ref_to_uuid(key)
            if uid in self._uuid_map:
                return next(k for k in self._uuid_map[uid])
        return None

    def _add_ext_ref_mapping(self, entity):
        if entity.external_ref in self._ext_ref_mapping:
            if entity == self[entity.external_ref]:
                return  # skip it
            raise EntityExists('External Ref %s already refers to %s' % (entity.external_ref, self[entity.external_ref]))
        self._ext_ref_mapping[entity.external_ref] = entity.link

    def __getitem__(self, item):
        """
        Note: this user-friendliness check adds 20% to the execution time of getitem-- so avoid it if possible
        (use _get_entity directly -- especially now that upstream is now deprecated)
        (note that _get_entity does not get contexts)

        :param item:
        :return:
        """
        if hasattr(item, 'link'):
            item = item.link
        return super(BasicArchive, self).__getitem__(item)

    @property
    def _archive_file(self):
        return os.path.join(self.source, 'entities.json')

    @property
    def _fragment_dir(self):
        return os.path.join(self.source, 'fragments')

    def __init__(self, fg_path, catalog=None, **kwargs):
        """

        :param fg_path:
        :param catalog: A foreground archive requires a catalog to deserialize saved fragments. If None, archive will
        still initialize (and will even be able to save fragments) but non-locally terminated fragments will fail.
        :param ns_uuid: Foreground archives may not use ns_uuids, so any namespace uuid provided will be ignored.
        :param kwargs:
        """
        if catalog is not None:
            kwargs['term_manager'] = kwargs.pop('term_manager', catalog.lcia_engine)
        super(LcForeground, self).__init__(fg_path, **kwargs)
        self._uuid_map = defaultdict(set)
        self._catalog = catalog
        self._ext_ref_mapping = dict()
        self._frags_with_flow = defaultdict(set)

        if not os.path.isdir(self.source):
            os.makedirs(self.source)
        self.load_all()
        self.check_counter('fragment')

    def catalog_ref(self, origin, external_ref, **kwargs):
        return self._catalog.catalog_ref(origin, external_ref, **kwargs)

    def _fetch(self, entity, **kwargs):
        return self.__getitem__(entity)

    def _load_all(self):
        if os.path.exists(self._archive_file):
            self._load_entities_json(self._archive_file)
            self._load_fragments()

    def make_interface(self, iface):
        if iface == 'foreground':
            return ForegroundImplementation(self)
        else:
            return super(LcForeground, self).make_interface(iface)

    def set_origin(self, origin):
        super(LcForeground, self).set_origin(origin)
        for k in self.entities_by_type('fragment'):
            k._origin = origin

    def _flow_ref_from_json(self, e, external_ref):
        origin = e.pop('origin')
        try:
            ref_qty_uu = e.pop('referenceQuantity')
        except KeyError:
            c = e.pop('characterizations')
            ref_qty_uu = next(cf['quantity'] for cf in c if 'isReference' in cf and cf['isReference'] is True)
        ref_qty = self[ref_qty_uu]
        ref = self.catalog_ref(origin, external_ref, entity_type='flow')
        if not ref.resolved:  # not found
            name = e.pop('Name', None) or 'unnamed flow %s' % origin
            ref = self.make_interface('foreground').add_or_retrieve(external_ref, ref_qty, name, **e)
        return ref

    def _make_entity(self, e, etype, ext_ref):
        if e['origin'] != self.ref:
            if etype == 'flow':
                return self._flow_ref_from_json(e, ext_ref)
            elif etype == 'quantity':
                return self.catalog_ref(e.pop('origin'), ext_ref, entity_type='quantity', **e)
        return super(LcForeground, self)._make_entity(e, etype, ext_ref)

    def add(self, entity):
        """
        Reimplement base add to (1) use link instead of external_ref, (2) merge instead of raising a key error.
        not sure we really want/need to do the merge thing- we will find out
        :param entity:
        :return:
        """
        if entity.origin is None:
            entity.origin = self.ref  # have to do this now in order to have the link properly defined
        elif entity.is_entity and entity.origin != self.ref:
            # TODO: Alert! entity properties are not preserved in the local ref
            entity = self.catalog_ref(entity.origin, entity.external_ref, entity_type=entity.entity_type)
            # for p in entity.properties:
            #     enew[p] = entity[p]  ...
        try:
            self._add(entity, entity.link)
        except EntityExists:
            if entity is self[entity.link]:
                pass
            elif entity.entity_type == 'fragment' and entity.external_ref != entity.uuid:
                raise ValueError('Name is already taken: %s' % entity.external_ref)
            else:
                # merge incoming entity's properties with existing entity
                current = self[entity.link]
                current.merge(entity)

        if hasattr(entity, 'uuid') and entity.uuid is not None:
            # self._entities[entity.uuid] = entity
            self._uuid_map[entity.uuid].add(entity.link)

        if entity.origin == self.ref and entity.external_ref != entity.uuid:
            self._add_ext_ref_mapping(entity)

        #TODO: figure out how to specify which flows should NOT be merged (current: only flows with no context)
        self._add_to_tm(entity)  # , merge_strategy='distinct')  # DWR!!! need to

        if entity.entity_type == 'fragment':
            self._frags_with_flow[entity.flow].add(entity)

    def _add_children(self, entity):
        if entity.entity_type == 'fragment':
            self.add_entity_and_children(entity.flow)
            for c in entity.child_flows:
                self.add_entity_and_children(c)
        else:
            super(LcForeground, self)._add_children(entity)

    def _rename_mechanics(self, frag, oldname):
        """
        This function updates all the various mappings between a fragment and its references,
        EXCEPT the ext_ref_mapping which is left to outside code because of complexities with sequencing

        This can surely be simplified (currently, renaming a fragment requires calling this twice) but not today.

        :param frag:
        :param oldname:
        :return:
        """
        self._entities[frag.link] = self._entities.pop(oldname)
        self._uuid_map[frag.uuid].remove(oldname)
        self._uuid_map[frag.uuid].add(frag.link)
        self._ents_by_type['fragment'].remove(oldname)
        self._ents_by_type['fragment'].add(frag.link)

    def _dename_fragment(self, prior):
        name = prior.external_ref
        if name == prior.uuid:
            return  # nothing to do
        print('removing name from fragment %s to %s' % (name, prior.uuid))
        priorlink = prior.link
        prior.de_name()
        self._ext_ref_mapping.pop(name)

        self._rename_mechanics(prior, priorlink)
        assert self._ref_to_key(name) is None

    def name_fragment(self, frag, name, auto=None, force=None):
        """
        This function is complicated because we have so many dicts:
         _entities maps link to entity
         _ext_ref_mapping maps custom name to link
         _uuid_map maps UUID to a set of links that share the uuid
         _ents_by_type keeps a set of links by entity type

        So, when we name a fragment, we want to do the following:
         - ensure the fragment is properly in the _entities dict
         - ensure the name is not already taken
         - set the fragment entity's name
         * pop the old link and replace its object with the new link in _entities
         * remove the old link and replace it with the new link in the uuid map
         * remove the old link and replace it with the new link in ents_by_type['fragment']
         * add the name with the new link to the ext_ref_mapping
        :param frag:
        :param name:
        :param auto: if True, if name is taken, apply an auto-incrementing numeric suffix until a free name is found
        :param force: if True, if name is taken, de-name the prior fragment and assign the name to the current one.
         This requires:
           - first de-naming the prior fragment,
           - then swapping its _entities entry
           - then swapping its _uuid_map entry
           - then swapping its _ents_by_type entry
           - then removing its duplicate ext_ref_mapping
           = then proceeding to normally rename the new frag
         If both auto and force are specified, auto takes precedence.
        :return: returns the assigned name
        """
        if frag.external_ref == name:
            return  # nothing to do-- fragment is already assigned that name
        current = self[frag.link]
        if current is not frag:
            if current is None:
                raise FragmentNotFound(frag)
            raise FragmentMismatch('%s\n%s' % (current, frag))
        if self._ref_to_key(name) is not None:
            if auto:
                inc = 0
                newname = '%s %02d' % (name, inc)
                while self._ref_to_key(newname) is not None:
                    inc += 1
                    newname = '%s %02d' % (name, inc)
                name = newname
            elif force:
                prior = self._entities[self._ref_to_key(name)]
                assert prior.external_ref == name
                self._dename_fragment(prior)
            else:
                raise ValueError('Name is already taken: "%s"' % name)
        oldname = frag.link
        try:
            frag.external_ref = name  # will raise PropertyExists if already set
        except PropertyExists:
            if force:
                self._dename_fragment(frag)
                oldname = frag.link
                frag.external_ref = name
            else:
                raise

        self._add_ext_ref_mapping(frag)
        self._rename_mechanics(frag, oldname)
        assert self._ref_to_key(name) == frag.link

        return name

    '''
    Save and load the archive
    '''
    def _do_load(self, fragments):
        for f in fragments:
            frag = LcFragment.from_json(self, f)
            # if frag.external_ref != frag.uuid:  # don't need this anymore- auto-added for entities with self.ref
            #     self._add_ext_ref_mapping(frag)
            self.add(frag)

        for f in fragments:
            frag = self[f['entityId']]
            try:
                frag.finish_json_load(self, f)
            except AttributeError:
                print(f)
                raise

    def _load_fragments(self):
        """
        This must be done in two steps, since fragments refer to other fragments in their definition.
        First step: create all fragments.
        Second step: set reference entities and terminations
        :return:
        """
        fragments = []
        if not os.path.exists(self._fragment_dir):
            os.makedirs(self._fragment_dir)
        for file in os.listdir(self._fragment_dir):
            if os.path.isdir(os.path.join(self._fragment_dir, file)):
                continue
            with open(os.path.join(self._fragment_dir, file), 'r') as fp:
                j = json.load(fp)

            fragments.extend(j['fragments'])
        self._do_load(fragments)

    def _recurse_frags(self, frag):
        frags = [frag]
        for x in sorted(frag.child_flows, key=lambda z: z.uuid):
            frags.extend(self._recurse_frags(x))
        return frags

    def save_fragments(self, save_unit_scores=False):
        current_files = os.listdir(self._fragment_dir)
        for r in self._fragments():
            frags = [t.serialize(save_unit_scores=save_unit_scores) for t in self._recurse_frags(r)]
            fname = r.uuid + '.json'
            if fname in current_files:
                current_files.remove(fname)
            tgt_file = os.path.join(self._fragment_dir, fname)
            with open(tgt_file, 'w') as fp:
                json.dump({'fragments': frags}, fp, indent=2, sort_keys=True)
        for leftover in current_files:
            if not os.path.isdir(os.path.join(self._fragment_dir, leftover)):
                print('deleting %s' % leftover)
                os.remove(os.path.join(self._fragment_dir, leftover))

    def save(self, save_unit_scores=False):
        self.write_to_file(self._archive_file, gzip=False, characterizations=True, values=True, domesticate=False)
        if not os.path.isdir(self._fragment_dir):
            os.makedirs(self._fragment_dir)
        self.save_fragments(save_unit_scores=save_unit_scores)

    def clear_unit_scores(self, lcia_method=None):
        for f in self.entities_by_type('fragment'):
            for s, t in f.terminations():
                if lcia_method is None:
                    t.clear_score_cache()
                else:
                    t.reset_score(lcia_method)
    '''
    Retrieve + display fragments
    '''
    def _fragments(self, background=None, **kwargs):
        for f in self.search('fragment', **kwargs):
            if f.reference_entity is None:
                if background is None or f.is_background == background:
                    yield f

    def _show_frag_children(self, frag, level=0, show=False):
        level += 1
        for k in frag.child_flows:
            if show:
                print('%s%s' % ('  ' * level, k))
            else:
                yield k
            for j in self._show_frag_children(k, level, show=show):
                yield j

    def fragments(self, *args, show_all=False, background=None, show=False, **kwargs):
        """
        :param : optional first param is filter string-- note: filters only on reference fragments!
        :param show_all: show child fragments as well as reference fragments
        :param background: [None] if True or False, show fragments whose background status is as specified
        :param show: [False] if true, print the fragments instead of returning them
        :param kwargs: search parameters
        :return:
        """
        for f in sorted([x for x in self._fragments(background=background, **kwargs)], key=lambda x: x.is_background):
            if len(args) != 0:
                if not bool(re.search(args[0], str(f), flags=re.IGNORECASE)):
                    continue
            if show:
                print('%s' % f)
                if show_all:
                    self._show_frag_children(f, show=show)
            else:
                yield f
                if show_all:
                    for k in self._show_frag_children(f):
                        yield k

    def fragments_with_flow(self, flow):
        for k in self._frags_with_flow[flow]:
            yield k

    def frag(self, string, many=False, strict=True):
        """
        strict=True is slow
        Works as an iterator. If nothing is found, raises StopIteration. If multiple hits are found and strict is set,
        raises Ambiguous Reference.
        :param string:
        :param many: [False] if true, generate
        :param strict: [True] synonym for not many
        :return:
        """
        if strict and not many:
            k = [f for f in self.fragments(show_all=True) if f.uuid.startswith(string.lower())]
            if len(k) > 1:
                for i in k:
                    print('%s' % i)
                raise AmbiguousReference(string)
            try:
                return k[0]
            except IndexError:
                raise FragmentNotFound(string)
        else:
            return self.frags(string)

    def frags(self, string):
        for f in self.fragments(show_all=True):
            if f.uuid.startswith(string.lower()):
                yield f

    def draw(self, string, **kwargs):
        if not isinstance(string, LcFragment):
            string = self.frag(string)
        string.show_tree(**kwargs)

    '''
    Utilities for finding terminated fragments and deleting fragments
    '''
    def delete_fragment(self, frag):
        """
        Need to purge the fragment from:
         _entities
         _ents_by_type
         _ext_ref_mapping
         _uuid_map
         _frags_with_flow
        and correct _counter.
        The fragment is not destroyed- and can be re-added. Deleting child fragments is left to interface code (why?)
        :param frag:
        :return:
        """
        if frag.reference_entity is not None:
            frag.unset_parent()
        self._entities.pop(frag.link)
        self._ents_by_type['fragment'].remove(frag.link)
        self._counter['fragment'] -= 1
        self._uuid_map.pop(frag.uuid)
        self._ext_ref_mapping.pop(frag.external_ref, None)
        self._frags_with_flow[frag.flow].remove(frag)

    def _del_f(self, f):
        print('Deleting %s' % f)
        del self._entities[f.uuid]

    def del_orphans(self, for_real=False):
        """
        self is a foreground archive -- delete
        """
        for f in self._fragments(background=True):
            if f.reference_entity is not None:
                continue
            try:
                next(self._find_links(f))
                print('Found a link for %s' % f)
            except StopIteration:
                print('### Found orphan %s' % f)
                if for_real:
                    self._del_f(f)

    def _find_links(self, frag):
        for i in self.fragments(show_all=True):
            for s, t in i.terminations():
                if t.term_node is frag:
                    yield t

    def linked_terms(self, frag):
        """
        returns a list of terminations that match the input.
        :param frag:
        :return:
        """
        return [f for f in self._find_links(frag)]



"""
I don't really want a foreground interface--- I want a foreground construction yard.  the query is not the right
route for this maybe?

I want to do the things in the editor: create and curate flows, terminate flows in processes, recurse on processes.

anything else?
child fragment flow generators and handlers. scenario tools.

scenarios currently live in the fragments- those need to be tracked somewhere. notably: on traversal, when parameters
are encountered.

Should that be here? how about monte carlo? where does that get done?

when do I go to bed?
"""

from antelope import ForegroundInterface


class ForegroundRequired(Exception):
    pass


_interface = 'foreground'


class AntelopeForegroundInterface(ForegroundInterface):
    """
    The bare minimum foreground interface allows a foreground to return terminations and to save anything it creates.
    """
    '''
    Left to subclasses
    '''
    def fragments(self, show_all=False, **kwargs):
        if show_all:
            raise ValueError('Cannot retrieve non-parent fragments via interface')
        for i in self._perform_query(_interface, 'fragments', ForegroundRequired,
                                     **kwargs):
            yield self.make_ref(i)

    def frag(self, string, many=False, **kwargs):
        """
        Return the unique fragment whose ID starts with string.

        Default: If the string is insufficiently specific (i.e. there are multiple matches), raise
        :param string:
        :param many: [False] if true, return a generator and don't raise an error
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'frag', ForegroundRequired,
                                   string, many=many, **kwargs)

    def new_fragment(self, flow, direction, **kwargs):
        """
        Create a fragment and add it to the foreground.

        If creating a child flow ('parent' kwarg is non-None), then supply the direction with respect to the parent
        fragment. Otherwise, supply the direction with respect to the newly created fragment.  Example: for a fragment
        for electricity production:

        >>> fg = ForegroundInterface(...)
        >>> elec = fg.new_flow('Electricity supply fragment', 'kWh')
        >>> my_frag = fg.new_fragment(elec, 'Output')  # flow is an output from my_frag
        >>> child = fg.new_fragment(elec, 'Input', parent=my_frag, balance=True)  # flow is an input to my_frag
        >>> child.terminate(elec_production_process)

        :param flow: a flow entity/ref, or an external_ref known to the foreground
        :param direction:
        :param kwargs: uuid=None, parent=None, comment=None, value=None, balance=False; **kwargs passed to LcFragment
        :return: the fragment? or a fragment ref? <== should only be used in the event of a non-local foreground
        """
        return self._perform_query(_interface, 'new_fragment', ForegroundRequired,
                                   flow, direction, **kwargs)

    def name_fragment(self, fragment, name, auto=None, force=None, **kwargs):
        """
        Assign a fragment a non-UUID external ref to facilitate its easy retrieval.  I suspect this should be
        constrained to reference fragments.  By default, if the requested name is taken, a ValueError is raised
        :param fragment:
        :param auto: if True, if name is taken, apply an auto-incrementing numeric suffix until a free name is found
        :param force: if True, if name is taken, de-name the prior fragment and assign the name to the current one
        :return:
        """
        return self._perform_query(_interface, 'name_fragment', ForegroundRequired,
                                   fragment, name, **kwargs)

    def fragments_with_flow(self, flow, direction=None, reference=None, background=None, **kwargs):
        """
        Generates fragments made with the specified flow, optionally filtering by direction, reference status, and
        background status.  For all three filters, the default None is to generate all fragments.
        :param flow:
        :param direction: [None | 'Input' | 'Output']
        :param reference: [None | False | True]
        :param background: [None | False | True]
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'fragments_with_flow', ForegroundRequired,
                                   flow, direction=direction, reference=reference, background=background, **kwargs)

    def find_or_create_term(self, exchange, background=None):
        """
        Finds a fragment that terminates the given exchange
        :param exchange:
        :param background: [None] - any frag; [True] - background frag; [False] - foreground frag
        :return:
        """
        return self._perform_query(_interface, 'find_or_create_term', ForegroundRequired,
                                   exchange, background=background)

    def create_fragment_from_node(self, process_ref, ref_flow=None, include_elementary=False):
        """

        :param process_ref: a ProcessRef
        :param ref_flow:
        :param include_elementary:
        :return:
        """
        return self._perform_query(_interface, 'create_fragment_from_node', ForegroundRequired,
                                   process_ref, ref_flow=ref_flow, include_elementary=include_elementary)

    def clone_fragment(self, frag, **kwargs):
        """

        :param frag: the fragment (and subfragments) to clone
        :param kwargs: suffix (default: ' (copy)', applied to Name of top-level fragment only)
                       comment (override existing Comment if present; applied to all)
        :return:
        """
        return self._perform_query(_interface, 'clone_fragment', ForegroundRequired,
                                   frag, **kwargs)

    def split_subfragment(self, fragment, replacement=None, **kwargs):
        """
                Given a non-reference fragment, split it off into a new reference fragment, and create a surrogate child
        that terminates to it.

        without replacement:
        Old:   ...parent-->fragment
        New:   ...parent-->surrogate#fragment;   (fragment)

        with replacement:
        Old:   ...parent-->fragment;  (replacement)
        New:   ...parent-->surrogate#replacement;  (fragment);  (replacement)

        :param fragment:
        :param replacement:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'split_subfragment', ForegroundRequired,
                                   fragment, replacement=replacement, **kwargs)

    def delete_fragment(self, fragment, **kwargs):
        """
        Remove the fragment and all its subfragments from the archive (they remain in memory)
        This does absolutely no safety checking.

        :param fragment:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'delete_fragment', ForegroundRequired,
                                   fragment, **kwargs)

    def observe(self, fragment, exchange_value=None, name=None, scenario=None, **kwargs):
        """
        Observe a fragment's exchange value with respect to its parent activity level.  Only applicable for
        non-balancing fragments whose parents are processes or foreground nodes (child flows of subfragments have
        their exchange values determined at traversal, as do balancing flows).

        A fragment should be named when it is observed.  This should replace name_fragment. In a completed model, all
        observable fragments should have names.

        Also use to define scenario exchange values (or to make replicate observations..).

        scenario and name should be mutually exclusive; if both are supplied, name is ignored.

        :param fragment:
        :param exchange_value:
        :param name:
        :param scenario:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'observe', ForegroundRequired,
                                   fragment, exchange_value, name=name, scenario=scenario, **kwargs)

    def set_balance_flow(self, fragment, **kwargs):
        """
        This should get folded into observe
        Specify that a given fragment is a balancing flow for the parent node, with respect to the specified fragment's
        flow's reference quantity.

        :param fragment:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'set_balance_flow', ForegroundRequired,
                                   fragment, **kwargs)

    def unset_balance_flow(self, fragment, **kwargs):
        """
        This should get folded into observe
        Specify that a given fragment's balance status should be removed.  The fragment's observed EV will remain at
        the most recently observed level.
        :param fragment:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'unset_balance_flow', ForegroundRequired,
                                   fragment, **kwargs)

    def create_process_model(self, process, ref_flow=None, set_background=True, **kwargs):
        """
        Create a fragment from a process_ref.  If process has only one reference exchange, it will be used automatically.
        By default, a child fragment is created for each exchange not terminated to context, and exchanges terminated
        to nodes are so terminated in the fragment.
        :param process:
        :param ref_flow: specify which exchange to use as a reference
        :param set_background: [True] whether to terminate the node to background
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'create_process_model', ForegroundRequired,
                                   process, ref_flow=ref_flow, **kwargs)

    def fragment_from_exchanges(self, exchanges, parent=None, include_context=False, multi_flow=False, **kwargs):
        """

        :param exchanges:
        :param parent: if parent is None, the first exchange is taken to be a reference fragment
        :param include_context: [False] if true, create subfragments terminating to context for elementary flows.
         otherwise leaves them unspecified (fragment LCIA includes unobserved exchanges)
        :param multi_flow:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'fragment_from_exchanges', ForegroundRequired,
                                   parent=parent, include_context=include_context,
                                   multi_flow=multi_flow, **kwargs)

    def traverse(self, fragment, scenario=None, **kwargs):
        """
        Traverse the fragment (observed) according to the scenario specification and return a list of FragmentFlows
        :param fragment:
        :param scenario:
        :param kwargs:
        :return:
        """
        return self._perform_query(_interface, 'traverse', ForegroundRequired,
                                   fragment, scenario, **kwargs)
